
#include "stdafx.h"
#include "Interpolator.h"
#include "FunctionTable.h"

float fPI = (float)PI;

float fadeLinear(float in) { return in; }
float fadeEqual(float in) { return sqrt(in); }
float fadeCosine(float in) { return cos(in*fPI)*-0.5f+0.5f; }
float fadeEqualCos1(float in) { return pow((cos((in+1.f)*fPI)*-0.5f+0.5f),1.3327765f)*-1.f+1.f; }
float fadeEqualCos2(float in) { return pow((cos((in+1.f)*fPI)*-0.5f+0.5f),1.771553f)*-1.f+1.f; }

FunctionTable equalTable = FunctionTable(fadeEqual);
FunctionTable cosineTable = FunctionTable(fadeCosine);
FunctionTable equalCos1Table = FunctionTable(fadeEqualCos1);
FunctionTable equalCos2Table = FunctionTable(fadeEqualCos2);

float lookupEqual(float in)		{ return equalTable.lookup(in); }
float lookupCosine(float in)	{ return cosineTable.lookup(in); }
float lookupEqualCos1(float in) { return equalCos1Table.lookup(in); }
float lookupEqualCos2(float in) { return equalCos2Table.lookup(in); }
/*
cos(((x-1)^2)*PI)*0.5+0.5
	very up-biased cosine

(cos(-(x-1)*PI)*-0.5+0.5)*(x-1)+1
	starts linear, ends cosine
*/

ptr_transform fades[] = {fadeLinear, lookupEqual, lookupCosine, lookupEqualCos1, lookupEqualCos2};
char *fadeNames[] = {"Linear", "Equal", "Cosine", "EqualCos1", "EqualCos2"};
const size_t numFades = std::extent<decltype(fades)>::value;
	
CMachineParameter const paraPosition = { pt_byte, "Position", "Position", 0, 254, 255, MPF_STATE, 0 };
CMachineParameter const paraInertia = { pt_byte, "Inertia", "Inertia", 0, 254, 255, MPF_STATE, 10 };
CMachineParameter const paraMode = { pt_byte, "Mode", "Mode", 0, numFades-1, 255, MPF_STATE, 3 };

static CMachineParameter const *pParameters[] = { 
	&paraPosition,
	&paraInertia,
	&paraMode
};

#pragma pack(1)

class gvals
{
public:
	byte position;
	byte inertia;
	byte mode;
};

#pragma pack()

CMachineInfo const MacInfo = 
{
	MT_EFFECT,						// type
	MI_VERSION,
	MIF_MULTI_IO,					// flags
	0,								// min tracks
	0,								// max tracks
	3,								// numGlobalParameters
	0,								// numTrackParameters
	pParameters,
	0, 
	NULL,
#ifdef KROSSIN
	"pac & MT KrossIn",
	"KrossIn",						// short name
#else
	"pac KrossOut",
	"KrossOut",						// short name
#endif
	"Pierre Cusa", 					// author
	"About..."
};

class mi;

class miex : public CMachineInterfaceEx
{
public:
	virtual void MultiWork(float const * const *inputs, float **outputs, int numsamples);
	virtual char const *GetChannelName(bool /*input*/, int index)
	{
		switch(index)
		{
#ifdef KROSSIN
		case 0: return "Signal A"; break;
		case 1: return "Signal B"; break;
#else
		case 0: return "Output A"; break;
		case 1: return "Output B"; break;
#endif
		default: return "<invalid>"; break;
		}
	}

public:
	mi *pmi;

};
class mi : public CMachineInterface
{
public:
	mi();
	virtual ~mi();

	virtual void Init(CMachineDataInput * const pi);
	virtual void Tick();
	virtual void Save(CMachineDataOutput * const po);
	char const* DescribeValue(int const param, int const v);
	virtual void Command(int const i);

	void MultiInWork(float const * const *inputs, float **outputs, int numsamples);
	void MultiOutWork(float const * const *inputs, float **outputs, int numsamples);

public:
	miex ex;
	gvals gval;
	Interpolator posInterpol;
	bool gotTick;
	int mode;
	//float previousTransA, previousTransB;
};

DLL_EXPORTS

void miex::MultiWork(float const * const *inputs, float **outputs, int numsamples) 
{ 
#ifdef KROSSIN
	pmi->MultiInWork(inputs, outputs, numsamples);
#else
	pmi->MultiOutWork(inputs, outputs, numsamples);
#endif
}


mi::mi()
{
	ex.pmi = this;
	GlobalVals = &gval;
	TrackVals = NULL;
	AttrVals = NULL;
	gotTick = false;
}

mi::~mi()
{
}

void mi::Init(CMachineDataInput * const pi)
{
	pCB->SetMachineInterfaceEx(&ex);
#ifdef KROSSIN
	pCB->SetInputChannelCount(2);
	pCB->SetOutputChannelCount(1);
#else
	pCB->SetInputChannelCount(1);
	pCB->SetOutputChannelCount(2);
#endif
}

void mi::Save(CMachineDataOutput * const po)
{
}

void mi::Tick()
{
	if (gval.position != paraPosition.NoValue){
		if (gotTick)
			posInterpol.setTarget((float)gval.position / (float)paraPosition.MaxValue);
		else {
			posInterpol.init((float)gval.position / (float)paraPosition.MaxValue);
			gotTick = true;
		}
	}

	if (gval.inertia != paraInertia.NoValue)
		posInterpol.inertia = (float)pMasterInfo->SamplesPerSec * 0.001f * (float)gval.inertia;

	if (gval.mode != paraMode.NoValue)
		mode = gval.mode;
}

void mi::MultiInWork(float const * const *inputs, float **outputs, int numsamples)
{
	if (inputs[0] == NULL && inputs[1] == NULL)
	{
		// setting an output pointer to NULL is equivalent of returning false in Work
		outputs[0] = NULL;
		return;
	}
	
	float2 * __restrict pinA = (float2 *)inputs[0];
	float2 * __restrict pinB = (float2 *)inputs[1];
	float2 * __restrict pout = (float2 *)outputs[0];

	ptr_transform transform = fades[mode];
	float transB = transform(posInterpol.value);
	float transA = transform(1.f - posInterpol.value);
	for (int i = 0; i < numsamples; i++) 
	{
		if (!posInterpol.advance())
		{
			// interpolation is in progress
			transB = transform(posInterpol.value);
			transA = transform(1.f - posInterpol.value);
		}
		pout[i] = float2(0,0);
		if (pinA) pout[i] += pinA[i]*transA;
		if (pinB) pout[i] += pinB[i]*transB;
	}
}

void mi::MultiOutWork(float const * const *inputs, float **outputs, int numsamples)
{
	if (inputs[0] == NULL)
	{
		// setting an output pointer to NULL is equivalent of returning false in Work
		for (int i = 0; i < 2; i++) outputs[i] = NULL;
		return;
	}

	float2 * __restrict pin = (float2 *)inputs[0];
	float2 * __restrict poutA = (float2 *)outputs[0];
	float2 * __restrict poutB = (float2 *)outputs[1];

	ptr_transform transform = fades[mode];
	float transB = transform(posInterpol.value);
	float transA = transform(1.f - posInterpol.value);
	for (int i = 0; i < numsamples; i++) 
	{
		if (!posInterpol.advance())
		{
			// interpolation is in progress
			transB = transform(posInterpol.value);
			transA = transform(1.f - posInterpol.value);
		}
		if (poutA) poutA[i] = pin[i]*transA;
		if (poutB) poutB[i] = pin[i]*transB;
	}
}

char const* mi::DescribeValue(int const param, int const v)
{
	static char txt[16];
	float temp;
	switch(param) {
	case 0:
		temp = (float)v/(float)paraPosition.MaxValue;
		sprintf(txt,"%.2f A - %.2f B", 1-temp, temp);
        return txt;
	case 1:
		temp = (float)v/100.f;
		sprintf(txt,"%.2f seconds", temp);
        return txt;
	case 2:
		sprintf(txt,"%s", fadeNames[v]);
        return txt;
	default:
		return NULL;
	}
}

void mi::Command(int const i)
{
	int j = 0;
    MessageBox(NULL,
		"     pac KrossIn - KrossOut v1.1\n\n     Inspired by MT Kross",
		"About",MB_OK|MB_SYSTEMMODAL);
}

